"""
This module implements the PipelineContainer (for compute and render pipelines).
This object is responsible for creating the native wgpu objects and doing the
actual dispatching / drawing.
"""

import wgpu

from ....utils import logger

from ..shader.base import ShaderInterface
from .utils import registry, GpuCache, hash_from_value
from .shared import get_shared
from .binding import Binding


# These caches enables sharing gpu resources for similar objects. It
# makes creating such objects faster (i.e. faster startup). It also
# saves gpu resources. It does not necessarily make the visualization
# faster.
LAYOUT_CACHE = GpuCache("layouts")
BINDING_CACHE = GpuCache("bindings")
SHADER_CACHE = GpuCache("shader_modules")
PIPELINE_CACHE = GpuCache("pipelines")


def get_cached_bind_group_layout(device, *args):
    key = "bind_group_layout", hash_from_value(args)
    result = LAYOUT_CACHE.get(key)
    if result is None:
        (entries,) = args
        result = device.create_bind_group_layout(entries=entries)

        LAYOUT_CACHE.set(key, result)

    return result


def get_cached_pipeline_layout(device, *args):
    key = "pipeline_layout", hash_from_value(args)
    result = LAYOUT_CACHE.get(key)
    if result is None:
        (bind_group_layouts,) = args
        result = device.create_pipeline_layout(bind_group_layouts=bind_group_layouts)

        LAYOUT_CACHE.set(key, result)

    return result


def get_cached_bind_group(device, *args):
    key = "bind_group", hash_from_value(args)
    result = BINDING_CACHE.get(key)
    if result is None:
        layout, entries = args
        result = device.create_bind_group(layout=layout, entries=entries)

        BINDING_CACHE.set(key, result)

    return result


def get_cached_shader_module(device, shader, shader_kwargs):
    # Using a key that *defines* the wgsl - rather than the wgsl itself
    # - avoids the templating to be applied on a cache hit, which saves
    # considerable time!
    key = "shader", shader.hash, hash_from_value(shader_kwargs)

    result = SHADER_CACHE.get(key)
    if result is None:
        wgsl = shader.generate_wgsl(**shader_kwargs)
        result = device.create_shader_module(code=wgsl)

        SHADER_CACHE.set(key, result)

    return result


def get_cached_compute_pipeline(device, *args):
    key = "compute_pipeline", hash_from_value(args)
    result = PIPELINE_CACHE.get(key)
    if result is None:
        pipeline_layout, shader_module = args

        result = device.create_compute_pipeline(
            layout=pipeline_layout,
            compute={"module": shader_module, "entry_point": "main"},
        )

        PIPELINE_CACHE.set(key, result)

    return result


def get_cached_render_pipeline(device, *args):
    key = "render_pipeline", hash_from_value(args)
    result = PIPELINE_CACHE.get(key)
    if result is None:
        (
            pipeline_layout,
            shader_module,
            primitive_topology,
            strip_index_format,
            cull_mode,
            depth_descriptor,
            color_descriptors,
        ) = args

        result = device.create_render_pipeline(
            layout=pipeline_layout,
            vertex={
                "module": shader_module,
                "entry_point": "vs_main",
                "buffers": [],
            },
            primitive={
                "topology": primitive_topology,
                "strip_index_format": strip_index_format,
                "front_face": wgpu.FrontFace.ccw,
                "cull_mode": cull_mode,
            },
            depth_stencil=depth_descriptor,
            multisample={
                "count": 1,
                "mask": 0xFFFFFFFF,
                "alpha_to_coverage_enabled": False,
            },
            fragment={
                "module": shader_module,
                "entry_point": "fs_main",
                "targets": color_descriptors,
            },
        )

        PIPELINE_CACHE.set(key, result)

    return result


def get_pipeline_container_group(wobject, environment):
    """Update the GPU objects associated with the given wobject. Returns
    quickly if no changes are needed. Only this function is used by the
    renderer.
    """

    # Get pipeline_container
    try:
        pipeline_container_group = wobject._wgpu_pipeline_container_group
    except AttributeError:
        pipeline_container_group = PipelineContainerGroup()
        wobject._wgpu_pipeline_container_group = pipeline_container_group
        changed_labels = {"create"}
    else:
        # Get whether the object has changes
        changed_labels = wobject.tracker.pop_changed()

    # Update if necessary - this path is defined to be fast if there are no changes
    # Don't put this under an ``if changed``, because work may be needed for a new environment.
    pipeline_container_group.update(wobject, environment, changed_labels)

    # Return the pipeline container group
    return pipeline_container_group


class PipelineContainerGroup:
    """This is a thin wrapper for a list of compute pipeline containers,
    and render pipeline containers. The purpose of this object is to
    obtain the appropriate shader objects and store them.
    """

    def __init__(self):
        self.environments_known = set()
        self.environments_uptodate = set()
        self.compute_containers = None
        self.render_containers = None
        self.bake_functions = None

    def update(self, wobject, environment, changed):
        """Update the pipeline containers that are wrapped. Creates (and re-creates)
        the containers if necessary.
        """

        if "create" in changed:
            self.compute_containers = ()
            self.render_containers = ()
            self.bake_functions = ()

            # Get render function for this world object,
            # and use it to get a high-level description of pipelines.
            renderfunc = registry.get_render_function(wobject)
            if renderfunc is None:
                raise ValueError(
                    f"Could not get a render function for {wobject.__class__.__name__} "
                    f"with {wobject.material.__class__.__name__}"
                )

            # Call render function
            with wobject.tracker.track_usage("create"):
                shaders = renderfunc(wobject)
                if isinstance(shaders, ShaderInterface):
                    shaders = [shaders]

            # Divide result over two bins, one for compute, and one for render
            compute_containers = []
            render_containers = []
            bake_functions = []
            for shader in shaders:
                assert isinstance(shader, ShaderInterface)
                if shader.type == "compute":
                    compute_containers.append(ComputePipelineContainer(shader))
                elif shader.type == "render":
                    render_containers.append(RenderPipelineContainer(shader))
                else:
                    raise ValueError(f"Shader type {shader.type} is unknown.")
                if shader.needs_bake_function:
                    bake_functions.append(shader.bake_function)

            # Store results
            self.compute_containers = tuple(compute_containers)
            self.render_containers = tuple(render_containers)
            self.bake_functions = tuple(bake_functions)

        # Manage known environments
        env_hash = environment.hash
        if env_hash not in self.environments_known:
            self.environments_known.add(env_hash)
            environment.register_pipeline_container(self)  # allows us to clean up

        # Update compute containers if something changed. Note that env_hash is "" for this path.
        if changed:
            for container in self.compute_containers:
                container.update(wobject, environment, "", changed)

        # If something has changed, and it was not updated for the current environment yet, update render pipelines now
        if changed:
            self.environments_uptodate.clear()
        if env_hash not in self.environments_uptodate:
            for container in self.render_containers:
                container.update(wobject, environment, env_hash, changed)

        # Mark this environment as up-to-date
        self.environments_uptodate.add(env_hash)

    def remove_env_hash(self, env_hash):
        """Called from the environment when it becomes inactive.
        This allows this object to remove all references to wgpu objects
        that won't be used, reclaiming memory on both CPU and GPU.
        """
        self.environments_known.discard(env_hash)
        self.environments_uptodate.discard(env_hash)
        for container in self.render_containers:
            container.remove_env_hash(env_hash)


class PipelineContainer:
    """The pipeline container stores the wgpu pipeline object as well as intermediate
    steps needed to create it. When an dependency of a certain step changes (which we track)
    then only the steps below it need to be re-run.
    """

    def __init__(self, shader: ShaderInterface):
        self.shader = shader
        self.shared = get_shared()  # the globally Shared object
        self.device = self.shared.device

        # Dict to store info on the wobject that affects shaders or pipeline.
        # Fields are set in a tracking-context to make sure things update accordingly.
        self.wobject_info = {}

        # The info that the shader generates
        self.shader_hash = b""
        self.bindings_dicts = None  # dict of dict of bindings
        self.pipeline_info = None  # dict
        self.render_info = None  # dict

        # The wgpu objects that we generate
        # These map env_hash to a list of objects (one for each pass).
        # For compute shaders the blend_mode is always "" and there is
        # one object in each.
        self.wgpu_shaders = {}
        self.wgpu_pipelines = {}

        # The bindings map group_index -> group
        self.bind_group_layout_descriptors = []
        self.wgpu_bind_group_layouts = []
        self.wgpu_bind_groups = []

        # A flag to indicate that an error occurred and we cannot dispatch
        self.broken = False

    def _check_bindings(self):
        assert isinstance(self.bindings_dicts, dict), "bindings_dicts must be a dict"
        for key1, bindings_dict in self.bindings_dicts.items():
            assert isinstance(key1, int), f"bindings slot must be int, not {key1}"
            assert isinstance(bindings_dict, dict), "bindings_dict must be a dict"
            for key2, b in bindings_dict.items():
                assert isinstance(key2, int), f"bind group slot must be int, not {key2}"
                assert isinstance(b, Binding), f"binding must be Binding, not {b}"

    def remove_env_hash(self, env_hash):
        """Called from the PipelineContainerGroup when an environment becomes inactive."""
        self.wgpu_shaders.pop(env_hash, None)
        self.wgpu_pipelines.pop(env_hash, None)

    def update(self, wobject, environment, env_hash, changed):
        """Make sure that the pipeline is up-to-date."""

        # Ensure that the information provided by the shader is up-to-date
        if changed:
            try:
                self.update_shader_data(wobject, changed)
            except Exception as err:
                self.broken = 1
                raise err
            else:
                self.broken = 0

        # Ensure that the (environment specific) wgpu objects are up-to-date
        if not self.broken:
            try:
                self.update_wgpu_data(environment, env_hash, changed)
            except Exception as err:
                self.broken = 2
                raise err
            else:
                self.broken = False

        if changed:
            logger.info(f"{wobject} shader update: {', '.join(sorted(changed))}.")

    def update_shader_data(self, wobject, changed):
        """Update the info that applies to all passes and environments."""

        if "create" in changed or "reset" in changed:
            with wobject.tracker.track_usage("reset"):
                self.wobject_info["pick_write"] = wobject.material.pick_write
            changed.update(("bindings", "pipeline_info", "render_info"))
            self.wgpu_shaders = {}

        if "bindings" in changed:
            self.shader.unlock_hash()
            with wobject.tracker.track_usage("!bindings"):
                self.bindings_dicts = self.shader.get_bindings(wobject, self.shared)
            self.shader.lock_hash()
            self._check_bindings()
            self.update_shader_hash()
            self.update_bind_groups()

        if "pipeline_info" in changed:
            with wobject.tracker.track_usage("pipeline_info"):
                self.pipeline_info = self.shader.get_pipeline_info(wobject, self.shared)
                self.wobject_info["depth_test"] = wobject.material.depth_test
            self._check_pipeline_info()
            changed.add("render_info")
            self.wgpu_pipelines = {}

        if "render_info" in changed:
            with wobject.tracker.track_usage("render_info"):
                self.render_info = self.shader.get_render_info(wobject, self.shared)
            self._check_render_info()

    def update_wgpu_data(self, environment, env_hash, changed):
        """Update the actual wgpu objects."""

        # Note: from here-on, we cannot access the wobject anymore, because any tracking should
        # be done in update_shader_data(). If you find that info on the wobject is needed,
        # add that info to self.wobject_info under the appropriate tracking context.

        # Determine what render-passes apply, for this combination of shader and blender
        if isinstance(self, RenderPipelineContainer):
            render_mask = self.render_info["render_mask"]
            blender = environment.blender
            pass_indices = []
            for pass_index in range(blender.get_pass_count()):
                if not render_mask & blender.passes[pass_index].render_mask:
                    continue
                if not blender.get_color_descriptors(
                    pass_index, self.wobject_info["pick_write"]
                ):
                    continue
                pass_indices.append(pass_index)
        else:
            pass_indices = [0]

        # Get shader objects for this environment
        env_shaders = self.wgpu_shaders.setdefault(env_hash, {})
        env_pipelines = self.wgpu_pipelines.setdefault(env_hash, {})

        # Update shaders
        for pass_index in pass_indices:
            if pass_index not in env_shaders:
                changed.add("compile_shader")
                env_shaders[pass_index] = self._compile_shader(pass_index, environment)
                env_pipelines.pop(pass_index, None)

        # Update pipelines
        for pass_index in pass_indices:
            if pass_index not in env_pipelines:
                changed.add("compose_pipeline")
                env_pipelines[pass_index] = self._compose_pipeline(
                    pass_index, environment, env_shaders
                )

    def update_shader_hash(self):
        """Update the shader hash, invalidating the wgpu shaders if it changed."""
        sh = self.shader.hash
        if sh != self.shader_hash:
            self.shader_hash = sh
            self.wgpu_shaders = {}

    def update_bind_groups(self):
        """
        - Calculate new bind_group_layout_descriptors (simple dicts).
        - When this has changed from our last version, we also update
          wgpu_bind_group_layouts and reset self.wgpu_pipelines
        - Calculate new wgpu_bind_groups.
        """

        # Check the bindings structure
        for i, bindings_dict in self.bindings_dicts.items():
            assert isinstance(i, int)
            assert isinstance(bindings_dict, dict)
            for j, b in bindings_dict.items():
                assert isinstance(j, int)
                assert isinstance(b, Binding)

        # Create two new dicts that correspond closely to the bindings.
        # Except this turns the dicts into lists.
        # These are the descriptors to create the wgpu bind groups and
        # bind group layouts.
        bg_descriptors = []
        bg_layout_descriptors = []
        for group_id in sorted(self.bindings_dicts.keys()):
            bindings_dict = self.bindings_dicts[group_id]
            while len(bg_descriptors) <= group_id:
                bg_descriptors.append([])
                bg_layout_descriptors.append([])
            bg_descriptor = bg_descriptors[group_id]
            bg_layout_descriptor = bg_layout_descriptors[group_id]
            for slot in sorted(bindings_dict.keys()):
                binding = bindings_dict[slot]
                binding_des, binding_layout_des = binding.get_bind_group_descriptors(
                    slot
                )
                bg_descriptor.append(binding_des)
                bg_layout_descriptor.append(binding_layout_des)

        # Clean up any trailing empty descriptors
        while bg_descriptors and not bg_descriptors[-1]:
            bg_descriptors.pop(-1)
            bg_layout_descriptors.pop(-1)

        # If the layout has changed, we need a new pipeline
        if self.bind_group_layout_descriptors != bg_layout_descriptors:
            self.bind_group_layout_descriptors = bg_layout_descriptors
            # Invalidate the pipeline
            self.wgpu_pipelines = {}
            # Create wgpu objects for the bind group layouts
            self.wgpu_bind_group_layouts = []
            for bg_layout_descriptor in bg_layout_descriptors:
                bind_group_layout = get_cached_bind_group_layout(
                    self.device, bg_layout_descriptor
                )
                self.wgpu_bind_group_layouts.append(bind_group_layout)

        # Always create wgpu objects for the bind groups. Note that this dict
        # includes the buffer texture objects.
        self.wgpu_bind_groups = []
        for bg_descriptor, layout in zip(bg_descriptors, self.wgpu_bind_group_layouts):
            bind_group = get_cached_bind_group(self.device, layout, bg_descriptor)
            self.wgpu_bind_groups.append(bind_group)


class ComputePipelineContainer(PipelineContainer):
    """Container for compute pipelines."""

    # These checks are here to validate the output of the shader
    # methods, not for user code. So its ok to use assertions here.

    def _check_pipeline_info(self):
        pipeline_info = self.pipeline_info
        assert isinstance(pipeline_info, dict)

        assert set(pipeline_info.keys()) == set(), f"{pipeline_info.keys()}"

    def _check_render_info(self):
        render_info = self.render_info
        assert isinstance(render_info, dict)

        assert set(render_info.keys()) == {"indices"}, f"{render_info.keys()}"

        indices = render_info["indices"]
        assert isinstance(indices, (tuple, list))
        assert all(isinstance(i, int) for i in indices)
        assert len(indices) == 3

    def _compile_shader(self, pass_index, env):
        """Compile the templateds wgsl shader to a wgpu shader module."""
        shader_kwargs = {}
        return get_cached_shader_module(self.device, self.shader, shader_kwargs)

    def _compose_pipeline(self, pass_index, env, shader_modules):
        """Create the wgpu pipeline object from the shader and bind group layouts."""

        # Create pipeline layout object from list of layouts
        pipeline_layout = get_cached_pipeline_layout(
            self.device, self.wgpu_bind_group_layouts
        )

        # Create pipeline object
        return get_cached_compute_pipeline(
            self.device, pipeline_layout, shader_modules[0]
        )

    def dispatch(self, compute_pass):
        """Dispatch the pipeline, doing the actual compute job."""
        if self.broken:
            return

        # Collect what's needed
        pipeline = self.wgpu_pipelines[""][0]
        indices = self.render_info["indices"]
        bind_groups = self.wgpu_bind_groups

        # Set pipeline and bindings
        compute_pass.set_pipeline(pipeline)
        for bind_group_id, bind_group in enumerate(bind_groups):
            compute_pass.set_bind_group(bind_group_id, bind_group, [], 0, 999999)

        # Compute!
        compute_pass.dispatch_workgroups(*indices)


class RenderPipelineContainer(PipelineContainer):
    """Container for render pipelines."""

    def __init__(self, shader):
        super().__init__(shader)
        self.strip_index_format = 0

    def _check_pipeline_info(self):
        pipeline_info = self.pipeline_info
        assert isinstance(pipeline_info, dict)

        expected = {"cull_mode", "primitive_topology"}
        assert set(pipeline_info.keys()) == expected, f"{pipeline_info.keys()}"
        self.update_strip_index_format()

    def _check_render_info(self):
        render_info = self.render_info
        assert isinstance(render_info, dict)

        expected = {"indices", "render_mask"}
        assert set(render_info.keys()) == expected, f"{render_info.keys()}"

        indices = render_info["indices"]
        assert isinstance(indices, (tuple, list))
        assert all(isinstance(i, int) for i in indices)
        assert len(indices) in (2, 4, 5)
        if len(indices) == 2:
            render_info["indices"] = indices[0], indices[1], 0, 0

        render_mask = render_info["render_mask"]
        assert isinstance(render_mask, int) and render_mask in (1, 2, 3)

    def update_strip_index_format(self):
        if not self.bindings_dicts or not self.pipeline_info:
            return
        # Set strip_index_format
        index_format = wgpu.IndexFormat.uint32
        strip_index_format = 0
        if "strip" in self.pipeline_info["primitive_topology"]:
            strip_index_format = index_format
        # Trigger a pipeline rebuild?
        if self.strip_index_format != strip_index_format:
            self.strip_index_format = strip_index_format
            self.wgpu_pipelines = {}

    def _compile_shader(self, pass_index, env):
        """Compile the templated shader to a list of wgpu shader modules
        (one for each pass of the blender).
        """
        blender = env.blender
        env_bind_group_index = len(self.wgpu_bind_groups)

        blender_kwargs = blender.get_shader_kwargs(pass_index)
        env_kwargs = env.get_shader_kwargs(env_bind_group_index)
        shader_kwargs = blender_kwargs.copy()
        shader_kwargs.update(env_kwargs)
        shader_kwargs["write_pick"] &= self.wobject_info["pick_write"]

        return get_cached_shader_module(self.device, self.shader, shader_kwargs)

    def _compose_pipeline(self, pass_index, env, shader_modules):
        """Create a list of wgpu pipeline objects from the shader, bind group
        layouts and other pipeline info (one for each pass of the blender).
        """

        strip_index_format = self.strip_index_format
        primitive_topology = self.pipeline_info["primitive_topology"]
        cull_mode = self.pipeline_info["cull_mode"]

        # Create pipeline layout object from list of layouts
        env_bind_group_layout, _ = env.wgpu_bind_group
        bind_group_layouts = [*self.wgpu_bind_group_layouts, env_bind_group_layout]

        pipeline_layout = get_cached_pipeline_layout(self.device, bind_group_layouts)

        # Instantiate the pipeline objects.
        # Note: The pipeline relies on the color and depth descriptors, which
        # include the texture format and a few other static things.
        # This step should *not* rerun when e.g. the canvas resizes.
        blender = env.blender
        depth_test = self.wobject_info["depth_test"]
        color_descriptors = blender.get_color_descriptors(
            pass_index, self.wobject_info["pick_write"]
        )
        depth_descriptor = blender.get_depth_descriptor(pass_index, depth_test)
        shader_module = shader_modules[pass_index]

        return get_cached_render_pipeline(
            self.device,
            pipeline_layout,
            shader_module,
            primitive_topology,
            strip_index_format,
            cull_mode,
            depth_descriptor,
            color_descriptors,
        )

    def draw(self, render_pass, environment, pass_index, render_mask):
        """Draw the pipeline, doing the actual rendering job."""
        if self.broken:
            return

        if not (render_mask & self.render_info["render_mask"]):
            return

        # Collect what's needed
        pipeline = self.wgpu_pipelines[environment.hash][pass_index]
        indices = self.render_info["indices"]
        bind_groups = self.wgpu_bind_groups

        # Set pipeline and bindings
        render_pass.set_pipeline(pipeline)

        for bind_group_id, bind_group in enumerate(bind_groups):
            render_pass.set_bind_group(bind_group_id, bind_group, [], 0, 99)

        env_bind_group_id = len(bind_groups)
        _, env_bind_group = environment.wgpu_bind_group
        render_pass.set_bind_group(env_bind_group_id, env_bind_group, [], 0, 99)

        # Draw!
        # draw(count_vertex, count_instance, first_vertex, first_instance)
        render_pass.draw(*indices)



"""
Defines the classes for the different render passes and the blender
objects that contain them. A blender becomes part of the environment
object.
"""

import wgpu  # only for flags/enums

from ....utils.enums import RenderMask
from .flusher import create_full_quad_pipeline
from .shared import get_shared

# Notes:
# - The user code provides color as-is in rgba.
# - In the get_fragment_output logic defined in the shaders here, the color
#   is pre-multiplied with the alpha.
# - All fixed-pipeling blending options assume premultiplied alpha.


standard_texture_des = {
    "sample_type": wgpu.TextureSampleType.unfilterable_float,
    "view_dimension": wgpu.TextureViewDimension.d2,
    "multisampled": False,
}


def blend_dict(src_factor, dst_factor, operation):
    return {
        "operation": operation,
        "src_factor": src_factor,
        "dst_factor": dst_factor,
    }


# %%%%%%%%%%  Define passes


class BasePass:
    """The base pass class, defining and documenting the API that a pass must provide."""

    render_mask = RenderMask.opaque | RenderMask.transparent
    write_pick = True

    def get_color_descriptors(self, blender, material_write_pick):
        """Get the list of fragment targets for device.create_render_pipeline()."""
        # The result affects the wobject's pipeline.
        # Returning [] prevents this pipeline from being created.
        return []

    def get_color_attachments(self, blender):
        """Get the list of color_attachments for command_encoder.begin_render_pass()."""
        # The result affects the rendering, but not the wobject's pipeline.
        # Returning [] prevents this pass from running.
        return []

    def get_depth_descriptor(self, blender):
        """Get dict that has the depth-specific fields for the depth_stencil field
        in device.create_render_pipeline().
        """
        return {}

    def get_depth_attachment(self, blender):
        """Get a dict that has the depth-specific fields for the
        depth_stencil_attachment argument of command_encoder.begin_render_pass().
        """
        return {}

    def get_shader_code(self, blender):
        """Get the fragment-write shader code.

        Notes:

        * This code gets injected into the shader, so the material shaders
          can use get_fragment_output.
        * This code should define FragmentOutput and get_fragment_output.
        """
        return ""


class OpaquePass(BasePass):
    """A pass that renders opaque fragments with depth testing, while
    discarting transparent fragments. This functions as the first pass
    in all multi-pass blenders.
    """

    render_mask = RenderMask.opaque
    write_pick = True

    def get_color_descriptors(self, blender, material_write_pick):
        bf, bo = wgpu.BlendFactor, wgpu.BlendOperation
        return [
            {
                "format": blender.color_format,
                "blend": {
                    "alpha": blend_dict(bf.one, bf.zero, bo.add),
                    "color": blend_dict(bf.one, bf.zero, bo.add),
                },
                "write_mask": wgpu.ColorWrite.ALL,
            },
            {
                "format": blender.pick_format,
                "blend": None,
                "write_mask": wgpu.ColorWrite.ALL if material_write_pick else 0,
            },
        ]

    def get_color_attachments(self, blender):
        color_load_op = pick_load_op = wgpu.LoadOp.load
        if blender.color_clear:
            blender.color_clear = False
            color_load_op = wgpu.LoadOp.clear
        if blender.pick_clear:
            blender.pick_clear = False
            pick_load_op = wgpu.LoadOp.clear

        return [
            {
                "view": blender.color_view,
                "resolve_target": None,
                "clear_value": (0, 0, 0, 0),
                "load_op": color_load_op,
                "store_op": wgpu.StoreOp.store,
            },
            {
                "view": blender.pick_view,
                "resolve_target": None,
                "clear_value": (0, 0, 0, 0),
                "load_op": pick_load_op,
                "store_op": wgpu.StoreOp.store,
            },
        ]

    def get_depth_descriptor(self, blender):
        return {
            "format": blender.depth_format,
            "depth_write_enabled": True,
            "depth_compare": wgpu.CompareFunction.less,
        }

    def get_depth_attachment(self, blender):
        depth_load_op = wgpu.LoadOp.load
        if blender.depth_clear:
            blender.depth_clear = False
            depth_load_op = wgpu.LoadOp.clear
        return {
            "view": blender.depth_view,
            "depth_clear_value": 1.0,
            "depth_load_op": depth_load_op,
            "depth_store_op": wgpu.StoreOp.store,
        }

    def get_shader_code(self, blender):
        return """
        struct FragmentOutput {
            @location(0) color: vec4<f32>,
            @location(1) pick: vec4<u32>,
        };
        fn get_fragment_output(depth: f32, color: vec4<f32>) -> FragmentOutput {
            if (color.a < 1.0 - ALPHA_COMPARE_EPSILON ) { discard; }
            var out : FragmentOutput;
            out.color = vec4<f32>(color.rgb, 1.0);
            return out;
        }
        """


class FullOpaquePass(OpaquePass):
    """A pass that considers all fragments opaque."""

    render_mask = RenderMask.opaque | RenderMask.transparent
    write_pick = True

    def get_shader_code(self, blender):
        return """
        struct FragmentOutput {
            @location(0) color: vec4<f32>,
            @location(1) pick: vec4<u32>,
        };
        fn get_fragment_output(depth: f32, color: vec4<f32>) -> FragmentOutput {
            var out : FragmentOutput;
            out.color = vec4<f32>(color.rgb, 1.0);  // always opaque
            return out;
        }
        """


class SimpleSinglePass(OpaquePass):
    """A pass that blends opaque and transparent fragments in a single pass."""

    render_mask = RenderMask.opaque | RenderMask.transparent
    write_pick = True

    def get_color_descriptors(self, blender, material_write_pick):
        bf, bo = wgpu.BlendFactor, wgpu.BlendOperation
        return [
            {
                "format": blender.color_format,
                "blend": {
                    "alpha": blend_dict(bf.one, bf.one_minus_src_alpha, bo.add),
                    "color": blend_dict(bf.one, bf.one_minus_src_alpha, bo.add),
                },
                "write_mask": wgpu.ColorWrite.ALL,
            },
            {
                "format": blender.pick_format,
                "blend": None,
                "write_mask": wgpu.ColorWrite.ALL if material_write_pick else 0,
            },
        ]

    def get_shader_code(self, blender):
        # Take depth into account, but don't treat transparent fragments differently
        return """
        struct FragmentOutput {
            @location(0) color: vec4<f32>,
            @location(1) pick: vec4<u32>,
        };
        fn get_fragment_output(depth: f32, color: vec4<f32>) -> FragmentOutput {
            var out : FragmentOutput;
            out.color = vec4<f32>(color.rgb * color.a, color.a);
            return out;
        }
        """


class SimpleTransparencyPass(BasePass):
    """A pass that only renders transparent fragments, blending them
    with the classic recursive alpha blending equation (a.k.a. the OVER
    operator).
    """

    render_mask = RenderMask.transparent
    write_pick = False

    def get_color_descriptors(self, blender, material_write_pick):
        bf, bo = wgpu.BlendFactor, wgpu.BlendOperation
        return [
            {
                "format": blender.color_format,
                "blend": {
                    "alpha": blend_dict(bf.one, bf.one_minus_src_alpha, bo.add),
                    "color": blend_dict(bf.one, bf.one_minus_src_alpha, bo.add),
                },
                "write_mask": wgpu.ColorWrite.ALL,
            },
        ]

    def get_color_attachments(self, blender):
        color_load_op = wgpu.LoadOp.load
        if blender.color_clear:
            blender.color_clear = False
            color_load_op = wgpu.LoadOp.clear

        return [
            {
                "view": blender.color_view,
                "resolve_target": None,
                "load_op": color_load_op,
                "store_op": wgpu.StoreOp.store,
            },
        ]

    def get_depth_descriptor(self, blender):
        return {
            "format": blender.depth_format,
            "depth_write_enabled": False,
            "depth_compare": wgpu.CompareFunction.less,
        }

    def get_depth_attachment(self, blender):
        depth_load_op = wgpu.LoadOp.load
        if blender.depth_clear:
            blender.depth_clear = False
            depth_load_op = wgpu.LoadOp.clear
        return {
            "view": blender.depth_view,
            "depth_load_op": depth_load_op,
            "depth_store_op": wgpu.StoreOp.store,
        }

    def get_shader_code(self, blender):
        return """
        struct FragmentOutput {
            @location(0) color: vec4<f32>,
        };
        fn get_fragment_output(depth: f32, color: vec4<f32>) -> FragmentOutput {
            if (color.a <= ALPHA_COMPARE_EPSILON) { discard; }
            var out : FragmentOutput;
            out.color = vec4<f32>(color.rgb * color.a, color.a);
            return out;
        }
        """


class WeightedTransparencyPass(BasePass):
    """A pass that implements weighted blended order-independent
    blending for transparent fragments, as proposed by McGuire in 2013.
    Multiple weight functions are supported.
    """

    render_mask = RenderMask.transparent
    write_pick = False

    def __init__(self, weight_func):
        if weight_func == "alpha":
            weight_code = """
                let weight = alpha;
            """
        elif weight_func == "depth":
            # The "generic purpose" weight function proposed by McGuire in
            # http://casual-effects.blogspot.com/2015/03/implemented-weighted-blended-order.html
            weight_code = """
                let a = min(1.0, alpha) * 8.0 + 0.01;
                let b = (1.0 - 0.99999 * depth);
                let weight = clamp(a * a * a * 1e8 * b * b * b, 1e-2, 3e2);
            """
        else:
            raise ValueError(
                f"Unknown WeightedTransparencyPass weight_func: {weight_func!r}"
            )

        self._weight_code = weight_code.strip()

    def get_color_descriptors(self, blender, material_write_pick):
        bf, bo = wgpu.BlendFactor, wgpu.BlendOperation
        return [
            {
                "format": blender.accum_format,
                "blend": {
                    "alpha": blend_dict(bf.one, bf.one, bo.add),
                    "color": blend_dict(bf.one, bf.one, bo.add),
                },
                "write_mask": wgpu.ColorWrite.ALL,
            },
            {
                "format": blender.reveal_format,
                "blend": {
                    "alpha": blend_dict(bf.zero, bf.one_minus_src_alpha, bo.add),
                    "color": blend_dict(bf.zero, bf.one_minus_src, bo.add),
                },
                "write_mask": wgpu.ColorWrite.ALL,
            },
        ]

    def get_color_attachments(self, blender):
        # We always clear the textures used in this pass
        accum_load_op = reveal_load_op = wgpu.LoadOp.clear
        accum_clear_value = 0, 0, 0, 0
        reveal_clear_value = 1, 0, 0, 0
        return [
            {
                "view": blender.accum_view,
                "resolve_target": None,
                "clear_value": accum_clear_value,
                "load_op": accum_load_op,
                "store_op": wgpu.StoreOp.store,
            },
            {
                "view": blender.reveal_view,
                "resolve_target": None,
                "clear_value": reveal_clear_value,
                "load_op": reveal_load_op,
                "store_op": wgpu.StoreOp.store,
            },
        ]

    def get_depth_descriptor(self, blender):
        return {
            "format": blender.depth_format,
            "depth_write_enabled": False,
            "depth_compare": wgpu.CompareFunction.less,
        }

    def get_depth_attachment(self, blender):
        # We never clear the depth buffer in this pass
        depth_load_op = wgpu.LoadOp.load
        return {
            "view": blender.depth_view,
            "depth_load_op": depth_load_op,
            "depth_store_op": wgpu.StoreOp.store,
        }

    def get_shader_code(self, blender):
        return """
        struct FragmentOutput {
            @location(0) accum: vec4<f32>,
            @location(1) reveal: f32,
        };
        fn get_fragment_output(depth: f32, color: vec4<f32>) -> FragmentOutput {
            let alpha = color.a;
            if (alpha <= ALPHA_COMPARE_EPSILON) { discard; }
            let premultiplied = color.rgb * alpha;
            WEIGHT_CODE
            var out : FragmentOutput;
            out.accum = vec4<f32>(premultiplied, alpha) * weight;
            out.reveal = alpha;  // yes, alpha, not weight
            return out;
            // Note 1: could also take user-specified transmittance into account.
            // Note 2: its also possible to undo a fragment contribution. For this the accum
            // and reveal buffer must be float to avoid clamping. And we'd do `abs(color.a)` above.
            // The formula would then be:
            //    out.accum = - out.accum;
            //    out.reveal = 1.0 - 1.0 / (1.0 - alpha);
        }
        """.replace(
            "WEIGHT_CODE", self._weight_code
        )


class FrontmostTransparencyPass(BasePass):
    """A render pass that renders the front-most transparent layer to
    a custom render target. This can then later be used in the combine-pass.
    """

    write_pick = True

    def get_color_descriptors(self, blender, material_write_pick):
        bf, bo = wgpu.BlendFactor, wgpu.BlendOperation
        return [
            {
                "format": blender.frontcolor_format,
                "blend": {
                    "alpha": blend_dict(bf.one, bf.zero, bo.add),
                    "color": blend_dict(bf.one, bf.zero, bo.add),
                },
                "write_mask": wgpu.ColorWrite.ALL,
            },
            {
                "format": blender.pick_format,
                "blend": None,
                "write_mask": wgpu.ColorWrite.ALL if material_write_pick else 0,
            },
        ]

    def get_color_attachments(self, blender):
        # We always clear the dedicated color textire, but we share the
        # pick texture, so we check the clear-flag.
        color_load_op = wgpu.LoadOp.clear
        pick_load_op = wgpu.LoadOp.load
        if blender.pick_clear:
            blender.pick_clear = False
            pick_load_op = wgpu.LoadOp.clear
        return [
            {
                "view": blender.frontcolor_view,
                "resolve_target": None,
                "clear_value": (0, 0, 0, 0),
                "load_op": color_load_op,
                "store_op": wgpu.StoreOp.store,
            },
            {
                "view": blender.pick_view,
                "resolve_target": None,
                "clear_value": (0, 0, 0, 0),
                "load_op": pick_load_op,
                "store_op": wgpu.StoreOp.store,
            },
        ]

    def get_depth_descriptor(self, blender):
        return {
            "format": blender.depth_format,
            "depth_write_enabled": True,
            "depth_compare": wgpu.CompareFunction.less,
        }

    def get_depth_attachment(self, blender):
        depth_load_op = wgpu.LoadOp.load
        if blender.depth_clear:
            blender.depth_clear = False
            depth_load_op = wgpu.LoadOp.clear
        return {
            "view": blender.depth_view,
            "depth_clear_value": 1.0,
            "depth_load_op": depth_load_op,
            "depth_store_op": wgpu.StoreOp.store,
        }

    def get_shader_code(self, blender):
        return """
        struct FragmentOutput {
            @location(0) color: vec4<f32>,
            @location(1) pick: vec4<u32>,
        };
        fn get_fragment_output(depth: f32, color: vec4<f32>) -> FragmentOutput {
            if (color.a <= ALPHA_COMPARE_EPSILON || color.a >= 1.0 - ALPHA_COMPARE_EPSILON) { discard; }
            var out : FragmentOutput;
            out.color = vec4<f32>(color.rgb * color.a, color.a);
            return out;
        }
        """


# %%%%%%%%%% Define blenders


class BaseFragmentBlender:
    """Manage how fragments are blended and end up in the final target.
    Each renderer has one blender object.
    """

    passes = []

    def __init__(self):
        self.device = get_shared().device

        # The size (2D in pixels) of the frame textures.
        self.size = (0, 0)

        # Objects for the combination pass
        self._combine_pass_pipeline = None
        self._combine_pass_bind_group = None

        # A dict that contains the metadata for all render targets.
        self._texture_info = {}

        # The below targets are always present, and the renderer expects their
        # format, texture, and view to be present.
        # These contribute to 4+4+8 = 16 bytes per pixel

        usg = wgpu.TextureUsage

        # The color texture is in srgb, because in the shaders we work with physical
        # values, but we want to store as srgb to make effective use of the available bits.
        self._texture_info["color"] = (
            wgpu.TextureFormat.rgba8unorm_srgb,
            usg.RENDER_ATTACHMENT | usg.COPY_SRC | usg.TEXTURE_BINDING,
        )

        # The depth buffer is 32 bit - we need that precision.
        # Note that there is also depth32float-stencil8, but it needs the
        # (webgpu) extension with the same name.
        self._texture_info["depth"] = (
            wgpu.TextureFormat.depth32float,
            usg.RENDER_ATTACHMENT | usg.COPY_SRC,
        )

        # The pick texture has 4 16bit channels, adding up to 64 bits.
        # These bits are divided over the pick data, e.g. 20 for the wobject id.
        self._texture_info["pick"] = (
            wgpu.TextureFormat.rgba16uint,
            usg.RENDER_ATTACHMENT | usg.COPY_SRC,
        )

    def clear(self):
        """Clear the buffers."""
        for key in self._texture_info.keys():
            setattr(self, key + "_clear", True)

    def clear_depth(self):
        """Clear the deph buffer only."""
        self.depth_clear = True

    def ensure_target_size(self, size):
        """If necessary, resize render-textures to match the target size."""

        assert len(size) == 2
        size = size[0], size[1]
        if size == self.size:
            return

        # Set new size
        self.size = size
        tex_size = size + (1,)

        # Any bind group is now invalid because they include source textures.
        self._combine_pass_bind_group = None

        # Recreate internal textures
        for name, (format, usage) in self._texture_info.items():
            wgpu_texture = self.device.create_texture(
                size=tex_size, usage=usage, dimension="2d", format=format
            )
            setattr(self, name + "_format", format)
            setattr(self, name + "_tex", wgpu_texture)
            setattr(self, name + "_view", wgpu_texture.create_view())
            setattr(self, name + "_clear", True)

    # The five methods below represent the API that the render system uses.

    def get_color_descriptors(self, pass_index, material_write_pick):
        return self.passes[pass_index].get_color_descriptors(self, material_write_pick)

    def get_color_attachments(self, pass_index):
        return self.passes[pass_index].get_color_attachments(self)

    def get_depth_descriptor(self, pass_index, depth_test=True):
        des = self.passes[pass_index].get_depth_descriptor(self)
        if not depth_test:
            des["depth_compare"] = wgpu.CompareFunction.always
            des["depth_write_enabled"] = False
        return {
            **des,
            "stencil_read_mask": 0,
            "stencil_write_mask": 0,
            "stencil_front": {},  # use defaults
            "stencil_back": {},  # use defaults
        }

    def get_depth_attachment(self, pass_index):
        return {
            **self.passes[pass_index].get_depth_attachment(self),
            "stencil_read_only": True,
            "stencil_load_op": wgpu.LoadOp.clear,
            "stencil_store_op": wgpu.StoreOp.discard,
        }

    def get_shader_kwargs(self, pass_index):
        return {
            "blending_code": self.passes[pass_index].get_shader_code(self),
            "write_pick": self.passes[pass_index].write_pick,
        }

    def get_pass_count(self):
        """Get the number of passes for this blender."""
        return len(self.passes)

    def perform_combine_pass(self):
        """Perform a render-pass to combine any multi-pass results, if needed."""

        # Get bindgroup and pipeline. The creation should only happens once per blender lifetime.
        if not self._combine_pass_pipeline:
            self._combine_pass_pipeline = self._create_combination_pipeline()
        if not self._combine_pass_pipeline:
            return []

        # Get the bind group. A new one is needed when the source textures resize.
        if not self._combine_pass_bind_group:
            self._combine_pass_bind_group = self._create_combination_bind_group(
                self._combine_pass_pipeline.get_bind_group_layout(0)
            )

        command_encoder = self.device.create_command_encoder()

        # Render
        render_pass = command_encoder.begin_render_pass(
            color_attachments=[
                {
                    "view": self.color_view,
                    "resolve_target": None,
                    "load_op": wgpu.LoadOp.load,
                    "store_op": wgpu.StoreOp.store,
                }
            ],
            depth_stencil_attachment=None,
            occlusion_query_set=None,
        )
        render_pass.set_pipeline(self._combine_pass_pipeline)
        render_pass.set_bind_group(0, self._combine_pass_bind_group, [], 0, 99)
        render_pass.draw(4, 1)
        render_pass.end()

        return [command_encoder.finish()]

    def _create_combination_pipeline(self):
        """Overload this to setup the specific combiner-pass."""
        return None


class OpaqueFragmentBlender(BaseFragmentBlender):
    """A fragment blender that pretends that all surfaces are opaque,
    even if they're not.
    """

    passes = [FullOpaquePass()]


class Ordered1FragmentBlender(BaseFragmentBlender):
    """A minimal fragment blender that uses the classic alpha blending
    equation, without differentiating between opaque and transparent
    objects. This is a common approach for applications using a single
    pass. Order dependent.
    """

    passes = [SimpleSinglePass()]


class Ordered2FragmentBlender(BaseFragmentBlender):
    """A first step towards better blending: separating the opaque from
    the transparent fragments, and blending the latter using the alpha
    blending equation. The second step has depth-testing, but no
    depth-writing. Order dependent.
    """

    passes = [OpaquePass(), SimpleTransparencyPass()]


class WeightedFragmentBlender(BaseFragmentBlender):
    """Weighted blended order independent transparency (McGuire 2013),
    using a weight function based only on alpha.

    The opaque pass is followed by as pass that accumulates the
    transparent fragments in two targets (4 channels) in a way that
    weights the fragments. These are combined in the combine-pass,
    realizing order independent blending.
    """

    passes = [OpaquePass(), WeightedTransparencyPass("alpha")]

    def __init__(self):
        super().__init__()

        usg = wgpu.TextureUsage

        # Create two additional render targets.
        # These contribute 8+1 = 9 bytes per pixel
        # So the total = 16 + 9 = 25 bytes per pixel

        # The accumulation buffer collects weighted fragments
        self._texture_info["accum"] = (
            wgpu.TextureFormat.rgba16float,
            usg.RENDER_ATTACHMENT | usg.TEXTURE_BINDING,
        )

        # The reveal buffer collects the weights.
        # McGuire: "Using R16F for the revealage render target will give slightly better
        # precision and make it easier to tune the algorithm, but a 2x savings on bandwidth
        # and memory footprint for that texture may make it worth compressing into R8 format."
        self._texture_info["reveal"] = (
            wgpu.TextureFormat.r8unorm,
            usg.RENDER_ATTACHMENT | usg.TEXTURE_BINDING,
        )

    def _create_combination_pipeline(self):
        binding_layout = [
            {
                "binding": 0,
                "visibility": wgpu.ShaderStage.FRAGMENT,
                "texture": standard_texture_des,
            },
            {
                "binding": 1,
                "visibility": wgpu.ShaderStage.FRAGMENT,
                "texture": standard_texture_des,
            },
        ]

        bf, bo = wgpu.BlendFactor, wgpu.BlendOperation
        targets = [
            {
                "format": self.color_format,
                "blend": {
                    "alpha": blend_dict(bf.one, bf.one_minus_src_alpha, bo.add),
                    "color": blend_dict(bf.one, bf.one_minus_src_alpha, bo.add),
                },
            },
        ]

        bindings_code = """
            @group(0) @binding(0)
            var r_accum: texture_2d<f32>;
            @group(0) @binding(1)
            var r_reveal: texture_2d<f32>;
        """

        fragment_code = """
            let epsilon = 1e-6;
            // Sample
            let accum = textureLoad(r_accum, texindex, 0).rgba;
            let reveal = textureLoad(r_reveal, texindex, 0).r;

            // Exit if no transparent fragments was written
            if (reveal >= 1.0) { discard; }  // no transparent fragments here

            // Reconstruct the color and alpha, and set final color, with premultiplied alpha
            let avg_color = accum.rgb / max(accum.a, epsilon);
            let alpha = 1.0 - reveal;
            out.color = vec4<f32>(avg_color * alpha, alpha);
        """

        return create_full_quad_pipeline(
            targets, binding_layout, bindings_code, fragment_code
        )

    def _create_combination_bind_group(self, bind_group_layout):
        # This must match the binding_layout above
        bind_group_entries = [
            {"binding": 0, "resource": self.accum_view},
            {"binding": 1, "resource": self.reveal_view},
        ]
        return self.device.create_bind_group(
            layout=bind_group_layout, entries=bind_group_entries
        )


class WeightedDepthFragmentBlender(WeightedFragmentBlender):
    """Weighted blended order independent transparency (McGuire 2013),
    using a general purpose depth weight function.
    """

    passes = [OpaquePass(), WeightedTransparencyPass("depth")]


class WeightedPlusFragmentBlender(WeightedFragmentBlender):
    """Three-pass weighted blended order independent transparency (McGuire 2013),
    using a weight function based on alpha, and in which the top-most
    transparent layer is actually in front.

    This uses the same approach as WeightedFragmentBlender, but in a
    3d pass we draw the front-most transparent layer. In the
    combine-pass, we subtract the front layer from the accum and reveal
    buffer, and add it again using the blend equation. In effect, the
    front-most layer is actually correct, and all transparent fragments
    behind it follow McGuire's approach. This looks a bit like a
    single-layer depth peeling.
    """

    # Since the front-most transparency pass also writes to the depth
    # buffer, we want to do the opaque pass first. And the weighed pass
    # should come ater the opaque pass.
    #
    # In an earlier version we did the front-most pass first, but then
    # each pass cleared the depth buffer. We should not do that if we
    # want to leave the depth buffer for post-processing or picking.
    #
    # Note that doing the front-post last means that the depth buffer
    # *does* include transparent fragments, while this is not the case
    # for the other blend modes.

    passes = [
        OpaquePass(),
        WeightedTransparencyPass("alpha"),
        FrontmostTransparencyPass(),
    ]

    def __init__(self):
        super().__init__()

        usg = wgpu.TextureUsage

        # Create one additional render target.
        # These contribute 4 bytes per pixel
        # So the total = 16 + 9 + 4 = 29 bytes per pixel

        # Color buffer for the front-most semitransparent layer
        self._texture_info["frontcolor"] = (
            wgpu.TextureFormat.rgba8unorm_srgb,
            usg.RENDER_ATTACHMENT | usg.TEXTURE_BINDING,
        )

    def _create_combination_pipeline(self):
        binding_layout = [
            {
                "binding": 0,
                "visibility": wgpu.ShaderStage.FRAGMENT,
                "texture": standard_texture_des,
            },
            {
                "binding": 1,
                "visibility": wgpu.ShaderStage.FRAGMENT,
                "texture": standard_texture_des,
            },
            {
                "binding": 2,
                "visibility": wgpu.ShaderStage.FRAGMENT,
                "texture": standard_texture_des,
            },
        ]

        bf, bo = wgpu.BlendFactor, wgpu.BlendOperation
        targets = [
            {
                "format": self.color_format,
                "blend": {
                    "alpha": blend_dict(bf.one, bf.one_minus_src_alpha, bo.add),
                    "color": blend_dict(bf.one, bf.one_minus_src_alpha, bo.add),
                },
            },
        ]

        bindings_code = """
            @group(0) @binding(0)
            var r_accum: texture_2d<f32>;
            @group(0) @binding(1)
            var r_reveal: texture_2d<f32>;
            @group(0) @binding(2)
            var r_frontcolor: texture_2d<f32>;
        """

        fragment_code = """
            let epsilon = 1e-6;

            // Sample
            var accum = textureLoad(r_accum, texindex, 0).rgba;
            var reveal = textureLoad(r_reveal, texindex, 0).r;
            let front = textureLoad(r_frontcolor, texindex, 0).rgba;

            // Exit if no transparent fragments was written - there also not be a front.
            if (reveal >= 1.0) { discard; }  // no transparent fragments here

            // Undo contribution of the front
            let weight = front.a;  // This must match with the other pass!
            accum = accum - front.rgba * weight;  // front is already pre-multiplied
            reveal = reveal / (1.0 - front.a);

            // Reconstruct the color and alpha, and set final color, with premultiplied alpha
            let avg_color = accum.rgb / max(accum.a, epsilon);
            let alpha = 1.0 - reveal;
            out.color = vec4<f32>(avg_color * alpha, alpha);

            // Blend in the front color (front is already premultiplied)
            let out_rgb = (1.0 - front.a) * out.color.rgb + front.rgb;
            let out_a = (1.0 - front.a) * out.color.a + front.a;
            out.color = vec4<f32>(out_rgb, out_a);
        """

        return create_full_quad_pipeline(
            targets, binding_layout, bindings_code, fragment_code
        )

    def _create_combination_bind_group(self, bind_group_layout):
        # This must match the binding_layout above
        bind_group_entries = [
            {"binding": 0, "resource": self.accum_view},
            {"binding": 1, "resource": self.reveal_view},
            {"binding": 2, "resource": self.frontcolor_view},
        ]
        return self.device.create_bind_group(
            layout=bind_group_layout, entries=bind_group_entries
        )


class AdditivePass(BasePass):
    """A pass that renders fragments with additive blending."""

    render_mask = RenderMask.opaque | RenderMask.transparent
    write_pick = False

    def get_color_descriptors(self, blender, material_write_pick):
        bf, bo = wgpu.BlendFactor, wgpu.BlendOperation
        return [
            {
                "format": blender.color_format,
                "blend": {
                    "alpha": blend_dict(bf.one, bf.one, bo.add),
                    "color": blend_dict(bf.one, bf.one, bo.add),
                },
                "write_mask": wgpu.ColorWrite.ALL,
            },
        ]

    def get_color_attachments(self, blender):
        color_load_op = wgpu.LoadOp.load
        if blender.color_clear:
            blender.color_clear = False
            color_load_op = wgpu.LoadOp.clear

        return [
            {
                "view": blender.color_view,
                "resolve_target": None,
                "load_op": color_load_op,
                "store_op": wgpu.StoreOp.store,
            },
        ]

    def get_depth_descriptor(self, blender):
        return {
            "format": blender.depth_format,
            "depth_write_enabled": False,
            "depth_compare": wgpu.CompareFunction.always,
        }

    def get_depth_attachment(self, blender):
        depth_load_op = wgpu.LoadOp.load
        if blender.depth_clear:
            blender.depth_clear = False
            depth_load_op = wgpu.LoadOp.clear
        return {
            "view": blender.depth_view,
            "depth_load_op": depth_load_op,
            "depth_store_op": wgpu.StoreOp.store,
        }

    def get_shader_code(self, blender):
        return """
        struct FragmentOutput {
            @location(0) color: vec4<f32>,
        };
        fn get_fragment_output(depth: f32, color: vec4<f32>) -> FragmentOutput {
            var out : FragmentOutput;
            out.color = color;
            return out;
        }
        """


class AdditiveFragmentBlender(BaseFragmentBlender):
    """A fragment blender that uses additive blending for all fragments."""

    passes = [AdditivePass()]



import wgpu  # only for flags/enums

from ....objects import Volume
from ....materials import VolumeSliceMaterial, VolumeRayMaterial
from ....resources import Texture

from .. import (
    register_wgpu_render_function,
    BaseShader,
    Binding,
    RenderMask,
    to_texture_format,
    GfxSampler,
    GfxTextureView,
    load_wgsl,
)

vertex_and_fragment = wgpu.ShaderStage.VERTEX | wgpu.ShaderStage.FRAGMENT


class BaseVolumeShader(BaseShader):
    def get_bindings(self, wobject, shared):
        geometry = wobject.geometry
        material = wobject.material  # noqa

        bindings = [
            Binding("u_stdinfo", "buffer/uniform", shared.uniform_buffer),
            Binding("u_wobject", "buffer/uniform", wobject.uniform_buffer),
            Binding("u_material", "buffer/uniform", material.uniform_buffer),
        ]

        # Collect texture and sampler
        if geometry.grid is None:
            raise ValueError("Volume.geometry must have a grid (texture).")
        if not isinstance(geometry.grid, Texture):
            raise TypeError("Volume.geometry.grid must be a Texture")
        if geometry.grid.dim != 3:
            raise TypeError("Volume.geometry.grid must a 3D texture (view)")

        tex_view = GfxTextureView(geometry.grid)
        sampler = GfxSampler(material.interpolation, "clamp")
        self["colorspace"] = geometry.grid.colorspace

        # Sampling type
        self["climcorrection"] = ""
        fmt = to_texture_format(geometry.grid.format)
        if "norm" in fmt or "float" in fmt:
            self["img_format"] = "f32"
            if "unorm" in fmt:
                self["climcorrection"] = " * 255.0"
            elif "snorm" in fmt:
                self["climcorrection"] = " * 255.0 - 128.0"
        elif "uint" in fmt:
            self["img_format"] = "u32"
        else:
            self["img_format"] = "i32"

        # Channels
        self["img_nchannels"] = len(fmt) - len(fmt.lstrip("rgba"))

        bindings.append(Binding("s_img", "sampler/filtering", sampler, "FRAGMENT"))
        bindings.append(Binding("t_img", "texture/auto", tex_view, vertex_and_fragment))

        # If a colormap is applied ...
        if material.map is not None:
            bindings.extend(
                self.define_img_colormap(material.map, material.map_interpolation)
            )
            self["colorspace"] = material.map.colorspace

        bindings = {i: b for i, b in enumerate(bindings)}
        self.define_bindings(0, bindings)

        return {
            0: bindings,
        }


@register_wgpu_render_function(Volume, VolumeSliceMaterial)
class VolumeSliceShader(BaseVolumeShader):
    type = "render"

    def get_pipeline_info(self, wobject, shared):
        return {
            "primitive_topology": wgpu.PrimitiveTopology.triangle_list,
            "cull_mode": wgpu.CullMode.none,
        }

    def get_render_info(self, wobject, shared):
        material = wobject.material

        render_mask = 0
        if wobject.render_mask:
            render_mask = wobject.render_mask
        elif material.is_transparent:
            render_mask = RenderMask.transparent
        else:
            # Determine what passes are needed
            if material.map is not None:
                if self["colormap_nchannels"] in (1, 3):
                    render_mask |= RenderMask.opaque
                else:
                    render_mask |= RenderMask.all
            else:
                if self["img_nchannels"] in (1, 3):
                    render_mask |= RenderMask.opaque
                else:
                    render_mask |= RenderMask.all

        return {
            "indices": (12, 1),
            "render_mask": render_mask,
        }

    def get_code(self):
        return load_wgsl("volume_slice.wgsl")


@register_wgpu_render_function(Volume, VolumeRayMaterial)
class VolumeRayShader(BaseVolumeShader):
    type = "render"

    def get_bindings(self, wobject, shared):
        self["mode"] = wobject.material.render_mode
        return super().get_bindings(wobject, shared)

    def get_pipeline_info(self, wobject, shared):
        return {
            "primitive_topology": wgpu.PrimitiveTopology.triangle_list,
            "cull_mode": wgpu.CullMode.front,  # the back planes are the ref
        }

    def get_render_info(self, wobject, shared):
        material = wobject.material

        render_mask = wobject.render_mask
        if not render_mask:
            if material.is_transparent:
                render_mask = RenderMask.transparent
            elif self["img_nchannels"] in (1, 3):
                render_mask = RenderMask.opaque
            else:
                render_mask = RenderMask.all

        return {
            "indices": (36, 1),
            "render_mask": render_mask,
        }

    def get_code(self):
        return load_wgsl("volume_ray.wgsl")



""" World Objects and Events.

.. currentmodule:: pygfx.objects

.. rubric:: Objects
.. autosummary::
    :toctree: objects/
    :template: ../_templates/custom_layout.rst

    WorldObject
    Group
    Scene
    Background
    Grid
    Points
    Line
    Mesh
    Image
    Volume
    Text
    Ruler
    InstancedMesh
    Light
    PointLight
    DirectionalLight
    AmbientLight
    SpotLight
    LightShadow
    DirectionalLightShadow
    SpotLightShadow
    PointLightShadow
    Bone
    Skeleton
    SkinnedMesh

.. rubric:: Events
.. autosummary::
    :toctree: objects/
    :template: ../_templates/custom_layout.rst

    Event
    EventTarget
    EventType
    PointerEvent
    KeyboardEvent
    RootEventHandler
    WheelEvent
    WindowEvent

"""

# flake8: noqa

from ._base import WorldObject
from ._events import (
    Event,
    EventTarget,
    EventType,
    PointerEvent,
    KeyboardEvent,
    RootEventHandler,
    WheelEvent,
    WindowEvent,
)
from ._more import (
    Group,
    Scene,
    Background,
    Grid,
    Points,
    Line,
    Mesh,
    Image,
    Volume,
    Text,
)
from ._ruler import Ruler
from ._instanced import InstancedMesh
from ._lights import Light, PointLight, DirectionalLight, AmbientLight, SpotLight
from ._lights import (
    LightShadow,
    DirectionalLightShadow,
    SpotLightShadow,
    PointLightShadow,
)
from ._skins import Bone, Skeleton, SkinnedMesh



from ._base import Material
from ..resources import Texture
from ..utils import unpack_bitfield, Color
from ..utils.enums import ColorMode, CoordSpace


class LineMaterial(Material):
    """Basic line material.

    Parameters
    ----------
    thickness : float
        The line thickness expressed in logical pixels. Default 2.0.
    thickness_space : str | CoordSpace
        The coordinate space in which the thickness is expressed ('screen', 'world', 'model'). Default 'screen'.
    color : Color
        The uniform color of the line (used depending on the ``color_mode``).
    color_mode : str | ColorMode
        The mode by which the line is coloured. Default 'auto'.
    map : Texture
        The texture map specifying the color for each texture coordinate. Optional.
    map_interpolation: str
        The method to interpolate the color map ('nearest' or 'linear'). Default 'linear'.
    dash_pattern : tuple
        The pattern of the dash, e.g. `[2, 3]`. See `dash_pattern` docs for details. Defaults to an empty tuple, i.e. no dashing.
    dash_offset : float
        The offset into the dash phase. Default 0.0.
    aa : bool
        Whether or not the line is anti-aliased in the shader. Default True.
    kwargs : Any
        Additional kwargs will be passed to the :class:`material base class <pygfx.Material>`.
    """

    uniform_type = dict(
        Material.uniform_type,
        color="4xf4",
        thickness="f4",
        dash_offset="f4",
    )

    def __init__(
        self,
        thickness=2.0,
        thickness_space="screen",
        *,
        color=(1, 1, 1, 1),
        color_mode="auto",
        map=None,
        map_interpolation="linear",
        dash_pattern=(),
        dash_offset=0,
        aa=True,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.thickness = thickness
        self.thickness_space = thickness_space
        self.color = color
        self.color_mode = color_mode
        self.map = map
        self.map_interpolation = map_interpolation
        self.dash_pattern = dash_pattern
        self.dash_offset = dash_offset
        self.aa = aa

    def _wgpu_get_pick_info(self, pick_value):
        # This should match with the shader
        values = unpack_bitfield(pick_value, wobject_id=20, index=26, coord=18)
        return {
            "vertex_index": values["index"],
            "segment_coord": (values["coord"] - 100000) / 100000.0,
        }

    @property
    def color(self):
        """The uniform color of the line."""
        return Color(self.uniform_buffer.data["color"])

    @color.setter
    def color(self, color):
        color = Color(color)
        self.uniform_buffer.data["color"] = color
        self.uniform_buffer.update_full()
        self._store.color_is_transparent = color.a < 1

    @property
    def color_is_transparent(self):
        """Whether the color is (semi) transparent (i.e. not fully opaque)."""
        return self._store.color_is_transparent

    @property
    def aa(self):
        """Whether the line's edges are anti-aliased.

        Aliasing gives prettier results by producing semi-transparent fragments
        at the edges. Lines thinner than one physical pixel are also diminished
        by making them more transparent.

        Note that by default, pygfx uses SSAA to anti-alias the total renderered
        result. Line-based aa results in additional improvement.

        Because semi-transparent fragments are introduced, it may affect how the
        line blends with other (semi-transparent) objects. It can also affect
        performance for very large datasets. In particular, when the line itself
        is opaque, the line is (in most blend modes) drawn twice to account for
        both the opaque and semi-transparent fragments.
        """
        return self._store.aa

    @aa.setter
    def aa(self, aa):
        self._store.aa = bool(aa)

    @property
    def color_mode(self):
        """The way that color is applied to the line.

        See :obj:`pygfx.utils.enums.ColorMode`:
        """
        return self._store.color_mode

    @color_mode.setter
    def color_mode(self, value):
        value = value or "auto"
        if value not in ColorMode:
            raise ValueError(
                f"LineMaterial.color_mode must be a string in {ColorMode}, not {repr(value)}"
            )
        self._store.color_mode = value

    @property
    def vertex_colors(self):
        return self.color_mode == ColorMode.vertex

    @vertex_colors.setter
    def vertex_colors(self, value):
        raise DeprecationWarning(
            "vertex_colors is deprecated, use ``color_mode='vertex'``"
        )

    @property
    def thickness(self):
        """The line thickness.

        The interpretation depends on `thickness_space`. By default it is in logical
        pixels, but it can also be in world or model coordinates.
        """
        return float(self.uniform_buffer.data["thickness"])

    @thickness.setter
    def thickness(self, thickness):
        self.uniform_buffer.data["thickness"] = max(0.0, float(thickness))
        self.uniform_buffer.update_full()

    @property
    def thickness_space(self):
        """The coordinate space in which the thickness (and dash_pattern) are expressed.

        See :obj:`pygfx.utils.enums.CoordSpace`:
        """
        return self._store.thickness_space

    @thickness_space.setter
    def thickness_space(self, value):
        value = value or "screen"
        if value not in CoordSpace:
            raise ValueError(
                f"LineMaterial.thickness_space must be a string in {CoordSpace}, not {repr(value)}"
            )
        self._store.thickness_space = value

    @property
    def map(self):
        """The texture map specifying the color for each texture coordinate.

        Can be None. The dimensionality of the map can be 1D, 2D or 3D, but
        should match the number of columns in the geometry's texcoords.
        """
        return self._store.map

    @map.setter
    def map(self, map):
        assert map is None or isinstance(map, Texture)
        self._store.map = map

    @property
    def map_interpolation(self):
        """The method to interpolate the colormap.

        Either 'nearest' or 'linear'.
        """
        return self._store.map_interpolation

    @map_interpolation.setter
    def map_interpolation(self, value):
        assert value in ("nearest", "linear")
        self._store.map_interpolation = value

    @property
    def dash_pattern(self):
        """The dash pattern.

        A sequence of floats describing the length of strokes and gaps. The
        length of the sequence must be an even number. Setting to None or the
        empty tuple means no dashing.

        For example, (5, 2, 1, 2) describes a a stroke of 5 units, a gap of 2,
        then a short stroke of 1, and another gap of 2. Units are relative to
        the line thickness (and therefore `thickness_space` also applies to  the
        `dash_pattern`).
        """
        return self._store.dash_pattern

    @dash_pattern.setter
    def dash_pattern(self, value):
        if value is None:
            value = ()
        if not isinstance(value, (tuple, list)):
            raise TypeError(
                "Line dash_pattern must be a sequence of floats, not '{value}'"
            )
        if len(value) % 2:
            raise ValueError("Line dash_pattern must have an even number of elements.")
        self._store.dash_pattern = tuple(max(0.0, float(v)) for v in value)

    @property
    def dash_offset(self):
        """The offset into the dash cycle to start drawing at, i.e. the phase."""
        return float(self.uniform_buffer.data["dash_offset"])

    @dash_offset.setter
    def dash_offset(self, value):
        self.uniform_buffer.data["dash_offset"] = float(value)
        self.uniform_buffer.update_full()


class LineDebugMaterial(LineMaterial):
    """Line debug material.

    A material that renders the triangles that the line is made up off.
    """

    pass


class LineSegmentMaterial(LineMaterial):
    """Line segment material.

    A material that renders line segments between each two subsequent points.
    """


class LineArrowMaterial(LineSegmentMaterial):
    """Arrow (vector) line material.

    A material that renders line segments that look like little arrows.
    """


class LineThinMaterial(LineMaterial):
    """Thin line material.

    A simple line, drawn with line_strip primitives that has a thickness
    of one physical pixel. Thickness, dashing, and aa are ignored.

    While you typically don't want to use this in your application (its
    width is inconsistent and looks *very* thin on HiDPI monitors), it can be
    useful for debugging as it is more performant than other line materials.

    """


class LineThinSegmentMaterial(LineMaterial):
    """Thin line segment material.

    Simple line segments, drawn with line primitives that has a thickness
    of one physical pixel. Thickness, dashing, and aa are ignored.

    While you typically don't want to use this in your application (its
    width is inconsistent and looks *very* thin on HiDPI monitors), it can be
    useful for debugging as it is more performant than other line materials.

    """



"""
The enums used in pygfx. The enums are all available from the root ``pygfx`` namespace.

.. currentmodule:: pygfx.utils.enums

.. autosummary::
    :toctree: utils/enums
    :template: ../_templates/custom_layout.rst

    ColorMode
    CoordSpace
    EdgeMode
    MarkerShape
    RenderMask
    SizeMode
    ElementFormat
    VisibleSide
    BindMode

"""

from wgpu.utils import BaseEnum


__all__ = [
    "RenderMask",
    "ColorMode",
    "EdgeMode",
    "SizeMode",
    "CoordSpace",
    "MarkerShape",
    "ElementFormat",
    "VisibleSide",
    "BindMode",
]


class Enum(BaseEnum):
    """Enum base class for pygfx."""


class RenderMask(Enum):
    """The RenderMask enum specifies the render passes in which an object participates."""

    auto = 0  #: Select the appropriate render passes automatically.
    opaque = 1  #: Only render in the opaque pass.
    transparent = 2  #: Only render in the transparancy pass.
    all = 3  #: Render in both passes.


class EdgeMode(Enum):
    centered = None  #: Centered edges (half the width on each side).
    inner = None  #: Inner edges (the width is added to the inside).
    outer = None  #: Outer edges (the width is added to the outside).


class ColorMode(Enum):
    """The ColorMode enum specifies how an object's color is established."""

    auto = None  #: Use either ``uniform`` and ``vertex_map``, depending on whether ``map`` is set.
    uniform = None  #: Use the uniform color (usually ``material.color``).
    vertex = None  #: Use the per-vertex color specified in the geometry  (usually  ``geometry.colors``).
    face = None  #: Use the per-face color specified in the geometry  (usually  ``geometry.colors``).
    vertex_map = None  #: Use per-vertex texture coords (``geometry.texcoords``), and sample these in ``material.map``.
    face_map = None  #: Use per-face texture coords (``geometry.texcoords``), and sample these in ``material.map``.
    debug = (
        None  #: Use colors most suitable for debugging. Defined on a per shader basis.
    )


class SizeMode(Enum):
    """The SizeMode enum specifies how an object's size/width/thickness is established."""

    uniform = None  #: Use a uniform size.
    vertex = None  #: Use a per-vertex size specified on the geometry.


class CoordSpace(Enum):
    """The CoordSpace enum specifies a coordinate space."""

    model = None  #: the WorldObject's own coordinate space.
    world = None  #: the coordinate space of the scene.
    screen = None  #: the coordiate space in logical pixels.


class MarkerShape(Enum):
    """The MarkerShape enum specifies the shape of a markers in the PointsMarkerMaterial."""

    circle = None  #: ● A circular shape (i.e. a disk).
    ring = None  #: ○ A circular shape with a hole in the middle.
    square = None  #: ■ A big square shape (sized to encompass the circle shape).
    diamond = None  #: ♦ A rotated square (sized to fit inside the circle).
    plus = None  #: + A plus symbol.
    cross = None  #: x A rotated plus symbol.
    asterix = None  #: ✳️ A plus and a cross combined.
    triangle_up = None  #: ▲
    triangle_down = None  #: ▼
    triangle_left = None  #: ◀
    triangle_right = None  #: ▶
    heart = None  #: ♥
    spade = None  #: ♠
    club = None  #: ♣
    pin = None  #: 📍
    custom = None  # Custom shape allowing users to provide their own SDF function


class ElementFormat(Enum):
    """The base elements to specify formats.

    These values can be used to compose formats of various layouts, e.g. 2D
    positions with "2xf4", rgb colors with "3xu8" or matrices with "4x4xf32".
    The purpose is to provide a common representation for simple formats, that
    can be used for buffers, textures and uniform buffers.
    """

    i1 = None  #: A signed 8bit integer.
    u1 = None  #: An unsigned 8-bit integer (byte).
    i2 = None  #: A signed 16-bit integer.
    u2 = None  #: An unsigned 16-bit integer.
    i4 = None  #: A signed 32-bit integer.
    u4 = None  #: An unsigned 32-bit integer.
    f2 = None  #: A 16-bit float.
    f4 = None  #: A 32-bit float.


class VisibleSide(Enum):
    """The VisibleSide enum specifies what side of a mesh is visible.

    Note that this is the inverse of the "CullMode", as it specifies what
    side is visible rather than what side is culled.
    """

    front = None  #: The front is visible.
    back = None  #: The back is visible.
    both = None  #: Both the front and back are visible.


class BindMode(Enum):
    """The BindMode enum specifies how a skinned mesh is bound to its skeleton."""

    attached = (
        "attached"  #: The skinned mesh shares the same world space as the skeleton.
    )
    detached = "detached"  #: The skinned mesh has its own world space.


# NOTE: Don't forget to add new enums to the toctree and __all__



import jinja2

loader = jinja2.PrefixLoader({}, delimiter=".")
loader.mapping["pygfx"] = jinja2.PackageLoader("pygfx.renderers.wgpu.wgsl", ".")

jinja_env = jinja2.Environment(
    block_start_string="{$",
    block_end_string="$}",
    variable_start_string="{{",
    variable_end_string="}}",
    line_statement_prefix="$$",
    undefined=jinja2.StrictUndefined,
    loader=loader,
)


def register_wgsl_loader(context, func):
    """Register a source for shader snippets.

    When code is encountered that looks like::

       {$ include 'some_context.name.wgsl' $}

    The loader for "some_context" is looked up and used to load the wgsl to include.
    This function allows registering a loader for your downstream package.

    Parameters
    ----------
    context : str
        The context of the loader.
    func: callable
        The function that will be called when a shader is loaded for the given context.
        The function must accept one positional argument (the name to include).
    """
    if not (isinstance(context, str) and "." not in context):
        raise TypeError("Wgsl load context must be a string witout dots.")
    if not callable(func):
        raise TypeError("The given wgsl load func must be callable.")
    if context in loader.mapping:
        raise RuntimeError(f"A loader is already registered for '{context}'.")
    loader.mapping[context] = func


def apply_templating(code, **kwargs):
    t = jinja_env.from_string(code)
    try:
        return t.render(**kwargs)
    except jinja2.UndefinedError as err:
        raise ValueError(f"Cannot compose shader: {err.args[0]}") from None



from math import pi, tan

import numpy as np
import pylinalg as la

from ..objects._base import WorldObject
from ..utils.transform import mat_inv
from ._base import Camera


class PerspectiveCamera(Camera):
    """A generic 3D camera with a configurable field of view (fov).

    Parameters
    ----------
    fov: float
        The field of view as an angle in degrees. Higher values give a
        wide-angle lens effect. This value is limited between 0 and
        179. If zero, it operates in orthographic mode. Default 50.
    aspect: float
        The desired aspect ratio, which is used to determine the vision pyramid's
        boundaries depending on the viewport size. Common values are 16/9 or 4/3. Default 1.
    width: float
        The width of the scene to view. If omitted or None, the width
        is derived from aspect and height.
    height: float
        The height of the scene to view. If omitted or None, the height
        is derived from aspect and width.
    zoom: float
        An additional zoom factor, equivalent to attaching a zoom lens.
    maintain_aspect: bool
        Whether the aspect ration is maintained as the window size changes.
        Default True. If false, the dimensions are stretched to fit the window.
    depth_range: 2-tuple
        The values for the near and far clipping planes. If not given
        or None, the clip planes will be calculated automatically based
        on the fov, width, and height.

    Note
    ----
    The width and/or height should be set when using a fov of zero,
    when you want to manipulate the camera with a controller, or when
    you want to make use of the automatic depth_range. However, if you
    also call ``show_pos``, ``show_object``, or ``show_rect`` you can omit
    width and height, because these methods set them for you.

    """

    _fov_range = 0, 179

    def __init__(
        self,
        fov=50,
        aspect=1,
        *,
        width=None,
        height=None,
        zoom=1,
        maintain_aspect=True,
        depth_range=None,
    ):
        super().__init__()

        self.fov = fov

        # Set width and height. Note that if both width and height are given, it overrides aspect
        aspect = aspect or 1
        if width is None and height is None:
            height = 100  # 100 produces safer near plane for most use-cases than 1
            width = height * aspect
        elif width is None:
            width = height * aspect
        elif height is None:
            height = width / aspect
        self.width = width
        self.height = height

        self.zoom = zoom
        self.maintain_aspect = maintain_aspect
        self.depth_range = depth_range

        self.set_view_size(1, 1)
        self.update_projection_matrix()

    @property
    def fov(self):
        """The field of view (in degrees), between 0-179."""
        return self._fov

    @fov.setter
    def fov(self, value):
        fov = float(value)
        fov = min(max(fov, self._fov_range[0]), self._fov_range[1])
        # Don't allow values between 0 and 1, as it becomes numerically unstable
        # For the record, zero is allowed, meaning orthographic projection
        if 0 < fov < 1:
            fov = 1
        self._fov = fov

    @property
    def width(self):
        """The (minimum) width of the view-cube.
        Changing the width changes the aspect, but not the height.
        """
        return self._width

    @width.setter
    def width(self, value):
        self._width = float(value)

    @property
    def height(self):
        """The (minimum) height of the view-cube.
        Changing the height changes the aspect, but not the width.
        """
        return self._height

    @height.setter
    def height(self, value):
        self._height = float(value)

    @property
    def aspect(self):
        """The aspect ratio (the ratio between width and height).
        Setting the aspect updates width and height such that their mean is unchanged.
        """
        return self._width / self._height

    @aspect.setter
    def aspect(self, value):
        aspect = float(value)
        if aspect <= 0:
            raise ValueError("aspect must be > 0")
        extent = 0.5 * (self._width + self._height)
        self._height = 2 * extent / (1 + aspect)
        self._width = self._height * aspect

    def _set_extent(self, extent):
        """Set the mean of width and height while maintaining aspect."""
        extent = float(extent)
        if extent <= 0:
            raise ValueError("extent must be > 0")
        aspect = self.aspect
        self._height = 2 * extent / (1 + aspect)
        self._width = self._height * aspect

    @property
    def zoom(self):
        """The camera zoom level."""
        return self._zoom

    @zoom.setter
    def zoom(self, value):
        self._zoom = float(value)

    @property
    def maintain_aspect(self):
        """Whether the aspect ration is maintained as the window size
        changes. Default True. Note that it only make sense to set this
        to False in combination with a panzoom controller.
        """
        return self._maintain_aspect

    @maintain_aspect.setter
    def maintain_aspect(self, value):
        self._maintain_aspect = bool(value)

    @property
    def depth_range(self):
        """The values for the near and far clip planes. If None, these values
        are calculated from fov, width, amd heiht.
        """
        return self._depth_range

    @depth_range.setter
    def depth_range(self, value):
        if value is None:
            self._depth_range = None
        elif isinstance(value, (tuple, list)) and len(value) == 2:
            self._depth_range = float(value[0]), float(value[1])
        else:
            raise TypeError("depth_range must be None or a 2-tuple.")

    def _get_near_and_far_plane(self):
        # Dept range explicitly given?
        if self._depth_range:
            return self._depth_range

        # Put 1000 units between the near and far plane, scaled by extent
        extent = 0.5 * (self._width + self._height)
        if self.fov > 0:
            # Scale near plane with the fov to compensate for the fact
            # that with very small fov you're probably looking at something
            # in the far distance.
            f = fov_distance_factor(self.fov)
            return (extent * f) / 1000, 1000 * extent
        else:
            return -500 * extent, 500 * extent

    @property
    def near(self) -> float:
        """The location of the near clip plane.
        Use `depth_range` so overload the computed value, if necessary.
        """
        near, _ = self._get_near_and_far_plane()
        return near

    @property
    def far(self) -> float:
        """The location of the far clip plane.
        Use `depth_range` so overload the computed value, if necessary.
        """
        _, far = self._get_near_and_far_plane()
        return far

    def get_state(self):
        """Get the state of the camera as a dict.

        The fields contain "position", "rotation", "scale", and
        "reference_up", representing the camera's transform. The scale
        is typically not used, but included for completeness. Further,
        the following properties are included: "fov", "width", "height",
        "zoom", "maintain_aspect", and "depth_range".

        """
        return {
            "position": self.local.position,
            "rotation": self.local.rotation,
            "scale": self.local.scale,
            "reference_up": self.world.reference_up,
            "fov": self.fov,
            "width": self.width,
            "height": self.height,
            "zoom": self.zoom,
            "maintain_aspect": self.maintain_aspect,
            "depth_range": self.depth_range,
        }

    def set_state(self, state):
        """Set the state of the camera from a dict.

        Accepted fields are the same as in ``get_state()``. In addition,
        the fields ``x``, ``y``, and ``z`` are also accepted to set the
        position along a singular dimension.

        """
        # Set the more complex props
        for key, value in state.items():
            if key == "position":
                self.local.position = value
            if key in ("x", "y", "z"):
                setattr(self.local, key, value)
            elif key == "scale":
                self.local.scale = value
            elif key == "rotation":
                self.local.rotation = value
            elif key == "reference_up":
                self.world.reference_up = value
            elif key in (
                "fov",
                "width",
                "height",
                "zoom",
                "maintain_aspect",
                "depth_range",
            ):
                # Simple props
                setattr(self, key, value)

    def set_view_size(self, width, height):
        self._view_aspect = width / height

    def update_projection_matrix(self):
        zoom_factor = self._zoom
        near, far = self._get_near_and_far_plane()

        if self.fov > 0:
            # Get the reference width / height
            size = 2 * near * tan(pi / 180 * 0.5 * self.fov) / zoom_factor
            # Pre-apply the reference aspect ratio
            height = 2 * size / (1 + self.aspect)
            width = height * self.aspect
            # Increase either the width or height, depending on the view size
            if not self._maintain_aspect:
                pass
            elif self.aspect < self._view_aspect:
                width *= self._view_aspect / self.aspect
            else:
                height *= self.aspect / self._view_aspect
            # Calculate bounds
            top = +0.5 * height
            bottom = -0.5 * height
            left = -0.5 * width
            right = +0.5 * width
            # Set matrices
            self.projection_matrix = la.mat_perspective(
                left, right, top, bottom, near, far, depth_range=(0, 1)
            )
            self.projection_matrix_inverse = mat_inv(self.projection_matrix)

        else:
            # The reference view plane is scaled with the zoom factor
            width = self.width / zoom_factor
            height = self.height / zoom_factor
            # Increase either the width or height, depending on the viewport shape
            aspect = width / height
            if not self._maintain_aspect:
                pass
            elif aspect < self._view_aspect:
                width *= self._view_aspect / aspect
            else:
                height *= aspect / self._view_aspect

            bottom = -0.5 * height
            top = +0.5 * height
            left = -0.5 * width
            right = +0.5 * width
            # Set matrices
            proj = la.mat_orthographic(
                left, right, top, bottom, near, far, depth_range=(0, 1)
            )
            proj_i = mat_inv(proj)
            self.projection_matrix = proj
            self.projection_matrix_inverse = proj_i

    def show_pos(self, target, *, up=None):
        """Look at the given position or object.

        This is similar to `look_at()`, but it also sets the width and
        height (while honoring aspect). So if you e.g. use an orbit
        controller on this camera, it rotates around the given target.

        Parameters
        ----------
        target: WorldObject or (x, y, z)
            The target to point the camera towards.
        up: 3-tuple
            If given, set ``camera.world.reference_up`` to the given value.

        """

        # Get pos from target
        if isinstance(target, WorldObject):
            pos = target.local.position
        else:
            pos = np.asarray(target)

        if pos.shape != (3,):
            raise ValueError("Expected position to have 3 values.")

        # Look at the provided position, taking up into account
        if up is not None:
            self.world.reference_up = up
        self.look_at(pos)

        # Update extent
        distance = la.vec_dist(pos, self.local.position)
        self._set_extent(distance / fov_distance_factor(self.fov))

    def show_object(self, target: WorldObject, view_dir=None, *, up=None, scale=1):
        """Orientate the camera such that the given target in is in view.

        Sets the position and rotation of the camera, and adjusts
        width and height to the target's size (while honoring aspect).

        This method is mainly intended for viewing 3D data. For 2D data
        it is not uncommon that the margins may feel somewhat large. Also
        see `show_rect()`.

        Parameters
        ----------
        target: WorldObject or a sphere (x, y, z, r)
            The object to look at.
        view_dir: 3-tuple of float
            Look at the object from this direction. If not given or None,
            uses the current view direction.
        up: 3-tuple
            If given, set ``camera.world.reference_up`` to the given value.
        scale: float
            Scale the size of what's shown. Default 1.

        """

        if up is None:
            up = self.world.reference_up
        else:
            up = np.asarray(up)
            self.world.reference_up = up

        # Get bounding sphere from target
        if isinstance(target, WorldObject):
            bsphere = target.get_world_bounding_sphere()
            if bsphere is None:
                raise ValueError(
                    "Given target does not have a bounding sphere, you should probably just provide a sphere (x, y, z, r) yourself."
                )
        elif isinstance(target, (tuple, list, np.ndarray)) and len(target) == 4:
            bsphere = tuple(target)
        else:
            raise TypeError(
                "show_object target must be a WorldObject, or a (x, y, z, radius) tuple."
            )

        # Obtain view direction
        if view_dir is None:
            rotation = self.local.rotation
            view_dir = la.vec_transform_quat((0, 0, -1), rotation)
        elif isinstance(view_dir, (tuple, list, np.ndarray)) and len(view_dir) == 3:
            view_dir = tuple(view_dir)
        else:
            raise TypeError(f"Expected view_dir to be sequence, not {view_dir}")
        view_dir = la.vec_normalize(view_dir)

        # Do the math ...
        view_pos = bsphere[:3]
        radius = max(0.0, bsphere[3]) or 1.0
        extent = radius * 2 * scale
        distance = fov_distance_factor(self.fov) * extent

        camera_pos = view_pos - view_dir * distance

        self.local.position = camera_pos
        self.look_at(view_pos)
        self._set_extent(extent)

    def show_rect(self, left, right, top, bottom, *, view_dir=None, up=None):
        """Position the camera such that the given rectangle is in view.

        The rectangle represents a plane in world coordinates, centered
        at the origin of the world, and rotated to be orthogonal to the
        view_dir.

        Sets the position and rotation of the camera, and adjusts
        width and height to the rectangle (thus also changing the aspect).

        This method is mainly intended for viewing 2D data, especially
        when `maintain_aspect` is set to False, and is convenient
        for setting the initial view before attaching a PanZoomController.

        Parameters
        ----------
        left: float
            The left boundary of the plane to show.
        right: float
            The right boundary of the plane to show.
        top: float
             The top boundary of the plane to show.
        bottom: float
             The bottom boundary of the plane to show.
        view_dir: 3-tuple of float
            Look at the rectangle from this direction. If not given or None,
            uses the current view direction.
        up: 3-tuple
           If given, set ``camera.world.reference_up`` to the given value.

        """

        if up is not None:
            self.world.reference_up = up

        # Obtain view direction
        if view_dir is None:
            view_dir = la.vec_transform_quat((0, 0, -1), self.world.rotation)
        else:
            view_dir = la.vec_normalize(view_dir)

        # Set bounds, note that this implicitly sets width, height (and aspect)
        self.width = right - left
        self.height = bottom - top
        extent = 0.5 * (self.width + self.height)
        # First move so we view towards the origin with the correct vector
        distance = fov_distance_factor(self.fov) * extent
        self.world.position = (0, 0, 0) - view_dir * distance
        self.look_at((0, 0, 0))

        # Now we have a rotation that we can use to orient our rect
        position = self.world.position
        rotation = self.world.rotation

        offset = 0.5 * (left + right), 0.5 * (top + bottom), 0
        new_position = position + la.vec_transform_quat(offset, rotation)
        self.world.position = new_position

    @property
    def frustum(self):
        """Corner positions of the viewing frustum in world space.

        Returns
        -------
        frustum : ndarray, [2, 4, 3]
            The coordinates of the frustum. The first axis corresponds to the
            frustum's plane (near, far), the second to the corner within the
            plane ((left, bottom), (right, bottom), (right, top), (left, top)),
            and the third to the world position of that corner.

        """

        projection_matrix = self.projection_matrix

        ndc_corners = np.array([(-1, -1), (1, -1), (1, 1), (-1, 1)])
        depths = np.array((0, 1))[:, None]
        local_corners = la.vec_unproject(ndc_corners, projection_matrix, depth=depths)
        world_corners = la.vec_transform(local_corners, self.world.matrix)
        return world_corners


def fov_distance_factor(fov):
    # It's important that controller and camera use the same distance calculations
    if fov > 0:
        fov_rad = fov * pi / 180
        factor = 0.5 / tan(0.5 * fov_rad)
    else:
        factor = 1.0
    return factor
